from funkyprompt.core import AbstractEntity, typing, Field, OpenAIEmbeddingField,RelationshipField
import datetime
from . import GenericEntityTypes
from pydantic import model_validator
from ast import literal_eval
from funkyprompt.core.utils.ids import funky_id
from funkyprompt.entities.relations import ProjectTask, TaskResource
from funkyprompt.core import utils

class Project(AbstractEntity):
    class Config:
        name: str = "project"
        namespace: str = "public"
        description: str = (
            """Projects allow people to manage what they care about, their goals etc. 
            It is possible to add and search projects and build relationships between projects and other entities"""
        )
 
    #TODO we need some context of user qualifier for key lookups and other queries. It could be an internal thing to the system
    #the agent does not need to know the user context but when saving AbstractUserEntity we could qualify the name as user/name 
    # and when resolving in he entity resolver we could "try" both keys i.e. the qualified one. For now for testing its ok to be single tenant
    name: str = Field(description="The unique name of the project")
    description: str = OpenAIEmbeddingField(
        description="The detailed description of the project"
    )
    target_completion: typing.Optional[datetime.datetime] = Field(
        default=None, description="An optional target completion date for the project"
    )
    labels: typing.Optional[typing.List[str] | str] = RelationshipField(
        default_factory=list,
        description="Optional category labels - should link to topic entities. When you are using labels you should always upsert or add labels to whatever is there already and never replace unless asked",
        entity_name=GenericEntityTypes.TOPIC,
    )
    
    @model_validator(mode="before")
    @classmethod
    def _types(cls, values):
        """we should be stricter in array/list types but here
        example of allowing lists as TEXT in stores
        """

        if isinstance(values.get("labels"), str):
            try:
                values["labels"] = literal_eval(values["labels"])
            except:
                pass

        return values


class Task(Project):
    class Config:
        name: str = "task"
        namespace: str = "public"
        description: str = (
            """Tasks allow people to manage small objectives as part of large projects. 
            It is possible to add and search tasks and build relationships between tasks and other entities"""
        )

    project_name: typing.Optional[str] = Field(
        default_factory=list,
        description="The associated project",
        entity_name=GenericEntityTypes.PROJECT,
    )

    status: typing.Optional[str] = Field(
        default="TODO",
        description="The status of the project e.g. TODO, DONE",
    )
    
    resource_names: typing.Optional[typing.List[str]] = Field(
        default_factory=list,
        description="A list of resources (unique name) that might be used in the task",
        entity_name=GenericEntityTypes.RESOURCE,
    )
    actionable: typing.Optional[str] = Field(default=None, description="A Low|Medium|High estimate of actionability")
    utility: typing.Optional[float] = Field(default=None, description="If the utility of the task can be estimated for the user's project or goals")
    effort: typing.Optional[float] = Field(default=None, description="An estimate of the difficulty of the task given what has been done so far")
   
    def get_relationships(cls):
        """
        instance method provides a list of edges defined on the object such as ProjectTasks
        instance methods are not accessible to agents
        """
        return None

    @model_validator(mode="before")
    @classmethod
    def _ids(cls, values):
        """tasks take ids based on their project and name
        it is up to the caller to ensure uniqueness
        """
        proj = values.get("project")
        name = f"{proj}/{values['name']}"
        values["id"] = funky_id(name)
        return values

    # TODO: im testing adding the inline task - but actually the agent should know this usually if we design things right (either the agent is Task or the planner provides the metadata)
    # TODO also testing moving crud to base class so that we can assume it on a type but using its schema and not the generic one in doc strings

    @classmethod
    def add(cls, task: "Task", **kwargs):
        """Save or update a task based on its task name as key

        #task model

        ```python
        class Task(BaseModel):
            name: str
            description: str
            project: Optional[str] = None
            labels: Optional[list[str]] = []
            target_completion: Optional[datetime]
        ```

        Args:
            task: The task object to add
        """
        from funkyprompt.services import entity_store

        if isinstance(task, dict):
            task = cls(**task)

        return entity_store(cls).update_records(task)

    @classmethod
    def set_task_status(cls, task_names: typing.List[str], status: str):
        """Move all tasks by name to the given status

        Args:
            task_names (typing.List[str]): list of one or more tasks for which to change status
            status (str): status as TODO or DONE
        """
        from funkyprompt.services import entity_store

        if task_names and not isinstance(task_names, list):
            task_names = [task_names]

        q = f"""UPDATE {cls.sql().table_name} set status=%s WHERE name = ANY(%s)"""

        return entity_store(cls).execute(q, (status, task_names))

    @classmethod
    def set_task_target_completion(
        cls, task_names: typing.List[str], date: str | datetime.datetime
    ):
        """Move all tasks by name to the given status

        Args:
            task_names (typing.List[str]): list of one or more tasks for which to change status
            date (str): the new date to complete the task
        """
        from funkyprompt.services import entity_store
        
        if task_names and not isinstance(task_names, list):
            task_names = [task_names]

        """could set these to upserts for cases where actually its not there"""
        
        q = f"""UPDATE {cls.sql().table_name} set target_completion=%s WHERE name = ANY(%s)"""

        return entity_store(cls).execute(q, (date, task_names))

    @classmethod
    def run_search(
        cls,
        questions: typing.List[str] | str,
        after_date: typing.Optional[dict] | str = None,
    ):
        """Query the tasks by natural language questions
        Args:
            questions (typing.List[str]|str): one or more questions to search for tasks
            date (str): the new date to complete the task
        """
        from funkyprompt.services import entity_store

        return entity_store(cls).ask(questions, after_date=after_date)

class Resource(AbstractEntity):
    class Config:
        name: str = "resource"
        namespace: str = "public"
        description: str = (
            """Resources are websites, data or people that can be involved in a project or task
               The may have unique domain names and they can be described
            """
        )
        
    uri: typing.Optional[str] = Field(description='a unique resource identifier if know')
    image_uri: typing.Optional[str] = Field(description='a representative image for the resource if known')
    category: str = Field(default=None,description="Resources can include IDEA|PERSON|WEBSITE|DATA|SOFTWARE and other categories")
    labels: typing.Optional[typing.List[str]] = RelationshipField(description='general labels to attach to the entity beyond category',default=None)
    
class TaskIdeaSummary(AbstractEntity):
    """this is a higher level example for testing the ideas
       the formatting of the type requires special care and then we will harden it in the pydantic type
       we should do a few different examples before we do because the model is very sensitive to the precise formatting of this
       however it should be something we can standardize for our scope
       we need something concise that expands child types
       also the dynamic data would be loaded based on some model or an entity lookup
    """
    class Config:
        name: str = 'task_idea_summary'
        namespace: str = 'public'
        description = """You are provided with a list of users prioritized goals and some data that could be useful.
Please summarize the main idea content and list resources in the form of entities and domains/websites.
Then produce a list of tasks as they relate to the users goals and projects.
"""
    
    """list resources | specific type -> people, companies, domain names, links, ideas"""
    content: str =  Field(default=None, description="Please summarize all the main useful ideas in the text")
        
    """lots of categorized resources - resource should also contain the category name so we can flat map db
       but we could override with an attribute if we really wanted to on our side - its just not clear what we want
       it might be an idea in this version to always be constrained to one child type entity and one child type relationships 
    """
    domain_names: typing.List[Resource] = Field(default_factory=list, description="a list of domain names, a category of resources")
    real_world_entities: typing.List[Resource] = Field(default_factory=list, description="a list of real world entities like people, websites, software, companies etc - these are a category of resources")
        
    """tasks which are a relationship type"""
    tasks: typing.List[Task] = Field(default_factory=dict, description="List tasks and the goal they map to. The goals of the user are listed and you suggest what actions they can take with respect to goals")
    
    @classmethod
    def _get_child_models(cls):
        """"""
        return [Resource]
    @classmethod
    def _get_prompting_data(cls):
        """
        this provides data injected into prompts for this type - can be dynamically loaded
        """
        return f"""
### User's prioritized goals
```text
1. create a business for personal knowledge management AI with rich support for databases of different types such as key-value, sql, vector and graph
2  write as much as possible and find good tools to manage my writing
3. understand the challenges people face today in terms of managing knowledge and personal growth
4. learn new AI methods and data modeling methods
5. build integrations from popular services and app
```
"""
    
    @classmethod
    def get_model_as_prompt(cls):
        """
        ---
        """
        #
        P = f"""
        
{cls._get_prompting_data()}

-----------------------------

_You will respond in Json using the following schema_

# Response schema

```python
class TaskIdeaSummary(BaseModel)
    id: typing.Union[str, uuid.UUID, NoneType] # A unique hash/uuid for the entity. The name can be hashed if its marked as the key
    name: <class 'str'> # The name is unique for the entity
    content: <class 'str'> # Please outline all the main useful ideas in the text in a lot of detail
    domain_names: typing.List[Resource] # a list of domain names, a category of resources
    real_world_entities: typing.List[Resource] # a list of real world entities like people, websites, software, companies etc - these are a category of resources
    tasks: typing.List[Task] # List tasks and the goal they map to. The goals of the user are listed and you suggest what actions they can take with respect to goals
```

#### This model uses child types below...

```python
class Resource(BaseModel)
    id: typing.Union[str, uuid.UUID, NoneType] # A unique hash/uuid for the entity. The name can be hashed if its marked as the key
    name: <class 'str'> # The name is unique for the entity
    url: typing.Optional[str] # the full url if given
    image_uri: typing.Optional[str] # a representative image for the resource if known
    category: <class 'str'> # Resources can include IDEA|PERSON|WEBSITE|DATA|SOFTWARE and other categories
    labels: typing.Optional[typing.List[str]] # general labels to attach to the entity beyond category
```


```python
class Task(BaseModel)
    name: <class 'str'> # The name is unique for the entity
    content: str # a description of the entity
    project: dict #the listed project or goal that this task would relate to - use the map of the `index` and `description` e.g. {{1:'my goal'}}
    utility_score: float # the possibly utility score with respect the users goals/project on a scale of 0 to 10
    actionability: str # the actionability of this task on a scale Low|Medium|High
    effort_days: int # the effort in days for this task
```

        """
        
        return P
    
    
"""form filling reference implementation
commonly we will want to collect data over time and its not obvious how to reliably do this without losing information over time. 
We rely on the reasoning ability of the agent and implement the pattern on faith for now
"""
class PersonPreferences(AbstractEntity):
    class Config:
        name: str = "person_preferences"
        namespace: str = "public"
        description: str = (
            """This is an example of a form filling pattern. The users preferences are updated over time. 
            Generally you should augment and extend the description of the person to create a nice overall summary that does not drop important information.
            Any concise facts that do not belong in the description could be added as attributes in the `misc_attributes` section if there is value to having a key property. Do not guess attributes and only added guessed attributes under the misc attributes section
            If you are given information about a user/person, you can update their details and preferences with this object.
            Entities are updated as type 4 slowly changing dimensions where we store and audit the user-agent conversation separately and we maintain current state in this object
            but its important not to overwrite any useful information but instead use an intelligent merge strategy.
            There are various fields that should be upserted for the user during conversation and there may be functions that can be called to save or lookup other details.
            It is important to consider the context of the user's preference when performing some tasks.
            You should be careful to unique identify a person. It is good to noramalzie names to title Case and observe ownership case e.g toms or tom's probably refers to the person Tom.
            """
        )

    name: typing.Optional[str] = Field(description="users name - good practice to always title case and remove any possession or plurals. For example toms or tom's should be mapped to Tom", default=None)
    email: typing.Optional[str] = Field(description='users email',default=None)
    description: typing.Optional[str] = OpenAIEmbeddingField(description='A detailed description of the person, their social network, interests etc. it is very important to retain all information that you gather and dont overwrite important details. Use an intelligence merge strategy', default=None)
    occupation: typing.Optional[str] = Field(description='persons occupation or role', default=None)
    favorite_topics: typing.Optional[typing.List[str]] = Field(default=None,description="A list of broad categories of interests the user has")
    date_of_birth: typing.Optional[datetime.date] = Field(default=None)
    life_goals:  typing.Optional[typing.List[str]] = Field(default=None,description="A list of medium to long term ambitions the user has")
    misc_attributes: typing.Optional[dict] = Field(default=None, description="Any concise factual attributes that do not fit neatly into an existing field e.g. the name of a pet. Dont put long form details here but instead add to description. Please supply a dict or something that can be parsed to a dict")
    related_entities: typing.Optional[dict] = RelationshipField(default=None, 
                                                                description="Any entities (by name) that are referenced can be replicated here with a short description of how they relate")
    
    @model_validator(mode="before")
    @classmethod
    def _types(cls, values):
        if values.get('life_goals'):
            values['life_goals'] = utils.coerce_list(values['life_goals'] )
        if values.get('favorite_topics'):
            values['favorite_topics'] = utils.coerce_list(values['favorite_topics'] )
        if values.get('misc_attributes'):
            values['misc_attributes'] = utils.coerce_json(values['misc_attributes'] )
        if values.get('related_entities'):
            values['related_entities'] = utils.coerce_json(values['related_entities'] )
        return values


import typing
from pydantic import BaseModel, Field, model_serializer, model_validator
from funkyprompt.core import AbstractModel, AbstractEntity
import datetime
import json
from funkyprompt.core.agents import CallingContext


class Message(BaseModel):
    role: str
    content: str | dict
    name: typing.Optional[str] = Field(
        description="for example function names", default=None
    )


class UserMessage(Message):
    role: str = "user"


class SystemMessage(Message):
    role: str = "system"


class MessageStack(BaseModel):
    """
    Message stack abstracts our notion of changing message context and provides some basic helpers
    Experiment with;
    - type prompts
    - known entities
    - smart pruning of context
    """

    model: AbstractModel | type = Field(
        description="The Model for the guidance of the agents - a model contains Config and functions and is required - it can be anything with a `get_model_description` method for this purpose"
    )

    current_date: typing.Optional[datetime.datetime] = Field(
        description="Id data is added to messages it can guide agents that are confused about the fourth dimension",
        default=None,
    )

    # could add base messages to but depending on what they are they could be typed things

    function_names: typing.Optional[typing.List[str]] = Field(
        description="Listing function names can be a useful hint", default_factory=list
    )

    question: typing.Optional[str] = Field(
        description="The users question", default=None
    )

    messages: typing.Optional[typing.List[Message]] = Field(
        description="A list of messages sent to the language mode context",
        default_factory=list,
    )

    language_model_provider: typing.Optional[str] = Field(
        description="The model provider can affect the serialization", default=None
    )

    known_entities: typing.Optional[typing.List[AbstractEntity]] = Field(
        default_factory=list,
        description="Known entities can be useful metadata if we encounter a small number of entities in the session - it costs context but if used can be powerful metadata",
    )

    @model_validator(mode="before")
    def _create(cls, values):
        """
        we basically want a prompt and a question but we can add in other stuff too
        """
        messages = values.get('messages') or []

        if not messages:
            """these are the default initial messages in the stack"""
            model = values.get("model")
            prompt = "Answer the users questions using 1. provided data, 2. any provided functions or 3. world knowledge depending on the context. Always check that you can use functions that you have. Do not use a search function if another function or existing data can be used in place. "
            
            #model description is more lightweight - there may be instances when its enough
            #e.g. the entity type and functions can be determine at runtime via lookups (TODO: optimize this)
            if hasattr(model, "get_model_as_prompt"):
                prompt += (model.get_model_as_prompt() or 'answer the users question\n')
            elif hasattr(model, "get_model_description"):
                prompt += (model.get_model_description() or 'answer the users question\n')
            """update messages from context, model and question"""

            messages.append(SystemMessage(content=prompt))
            date = values.get("current_date")
            if date:
                """this is added because sometimes it screws up date based queries"""
                messages.append(
                    SystemMessage(
                        content=f"I observe the current date is {date} so I should take that into account if asked questions about time"
                    )
                )

            function_names = values.get("function_names")
            if function_names:
                """this is added because sometimes it seems to need this nudge (TODO:)"""
                messages.append(
                    UserMessage(
                        content=f"""You can use the following functions by default {function_names} and in some cases you may be able to search and load others. 
                        If you encounter a new function by name that you do not have, you can always ask for it to be activated and you
                        should do that without asking the user, calling it where possible."""
                    )
                )

            """finally add the users question"""
            if values.get("question"):
                messages.append(UserMessage(content=values.get("question")))

        values["messages"] = messages

        return values

    def reset(cls):
        """returns the message stack only with the model elements"""
        data = dict(vars(cls))
        data["messages"] = []
        return MessageStack(model=cls.model, current_date=cls.current_date)

    def add(cls, message: Message | dict):
        """augment the message and retain all the other stuff"""
        if isinstance(message, dict):
            message = Message(**message)
        messages = cls.messages
        messages.append(message)
        data = dict(vars(cls))
        data["messages"] = messages

        return MessageStack(**data)

    def add_system_message(cls, data: str):
        """add string or dict content as system message"""
        return cls.add(SystemMessage(content=data))

    def add_user_message(cls, data: str | dict):
        """add string or dict content as system message"""
        return cls.add(UserMessage(content=data))

    @classmethod
    def structure_question(
        cls, question: str, model: AbstractModel, context: CallingContext = None
    ) -> typing.List[dict]:
        """Prompt building in `funkyprompt` is governed by models or types. A question is injected into a scaffold based on the type
        For language models we can construct a set of one or more messages using different roles.
        One large blob in a question can work but its often sensible to split fragments by role

        Args:
            question (str): the users question
            model (AbstractModel): the model used to generate the plan
            context (CallingContext, optional): context is used to determine the model provider and other session context. Defaults to None.
        """
        return []
    
    @classmethod
    def from_q_and_a(cls, question:str, data: typing.Any, observer_model:AbstractModel=AbstractEntity):
        """
        often we will have a question that will trigger a search
        the search result can then be 'observed' by an observer model that responds to the user
        the result of this is typically to provide a formatted response to the user
        
        the data should be serializable to json
        
        """
        messages= [
            SystemMessage(content=f'The data below were retrieved from a resource and can be used to answer the users question\n````json{json.dumps(data,default=str)}``')
        ]
        return MessageStack(question=question,messages=messages, model=observer_model)
    
    @classmethod
    def format_function_response_data(
        cls, name: str, data: typing.Any, context: CallingContext = None
    ) -> Message:
        """format the function response for the agent - essentially just a json dump

        Args:
            name (str): the name of the function
            data (typing.Any): the function response
            context (CallingContext, optional): context such as what model we are using to format the message with

        Returns: formatted messages for agent as a dict
        """
        
        """Pydantic things """
        if hasattr(data,'model_dump'):
            data = data.model_dump()

        return Message(
            role="function",
            name=f"{str(name)}",
            content=json.dumps(
                {
                    # do we need to be this paranoid most of the time?? this is a good label to point later stages to the results
                    "about-these-data": "here are some data that may or may not contain the answer to your question - please review it carefully",
                    "data": data,
                },
                default=str,
            ),
        )

    @classmethod
    def format_function_response_type_error(
        cls, name: str, ex: Exception, context: CallingContext = None
    ) -> Message:
        """type errors imply the function was incorrectly called and the agent should try again

        Args:
            name (str): the name of the function
            data (typing.Any): the function response
            context (CallingContext, optional): context such as what model we are using to format the message with

        Returns: formatted error messages for agent as a dict
        """
        return Message(
            role="system",
            name=f"{str(name.replace('.','_'))}",
            content=f"""You have called the function incorrectly - try again {ex}""",
        )

    def format_function_response_error(
        name: str, ex: Exception, context: CallingContext = None
    ) -> Message:
        """general errors imply something wrong with the function call

        Args:
            name (str): the name of the function
            data (typing.Any): the function response
            context (CallingContext, optional): context such as what model we are using to format the message with

        Returns: formatted error messages for agent as a dict
        """

        return Message(
            role="system",
            name=f"{str(name.replace('.','_'))}",
            content=f"""This function failed - you should try different arguments or a different function. - {ex}. 
If no data are found you must search for another function if you can to answer the users question. 
Otherwise check the error and consider your input parameters """,
        )

    """smart pruning of messages"""

    @model_serializer()
    def custom_serializer(self):
        """
        we can be smarter about this
        """

        return [m.model_dump() for m in self.messages]



from funkyprompt.core import AbstractModel
import typing
from funkyprompt.services.data import DataServiceBase
from funkyprompt.core.utils import logger, env
from funkyprompt.core.types.sql import VectorSearchOperator
import os
from lancedb.table import LanceDataset, LanceTable
import pyarrow as pa
from lancedb.pydantic import LanceModel, Vector
from lancedb.embeddings import get_registry
from funkyprompt.services import fs
"""TODO env.preferred_test_embedding"""
func = get_registry().get("openai").create(name="text-embedding-ada-002")


class LanceAbstractContentModel(LanceModel, AbstractModel):
    """
    The abstract content model extends the AbstractModel to focus on unstructured data that we might want embeddings for (text,image)

    MyModel = AbstractContentModel(name='test', content='test', vector=nd.zeros(EmbeddingFunctions.openai.ndims()))
    See
    """

    #TODO - have the pyarrow support the union type
    id: typing.Optional[str] = None
    vector: Vector(func.ndims()) = func.VectorField()
    content: str = func.SourceField()
    name: str
    
class LanceDBService(DataServiceBase):
    """the duckdb sql model uses lancedb as an assistant vector index"""

    def _lance_connect(self, use_async=True):
        """
        This is basically for S3 access and we are specific about it
        """
        import lancedb
        #root = _setup(root)
 
        return lancedb.connect(self._db_uri)#region=os.environ.get("AWS_DEFAULT_REGION")
  
    def __init__(self, model: AbstractModel, auto_register:bool=True, schema: pa.Schema = None):
        """the model is going to be mapped to the Lance abstract content model to take advantage of its auto-vector bits
           conveniently this only needs to be in the construction and we can write dicts to the db later
        """
        self.model =  LanceAbstractContentModel.create_model(name=model.get_model_name(), 
                                           namespace=model.get_model_namespace(), 
                                                #fields=model.model_fields
                                                )
        
        self._db_uri = f"{env.STORE_ROOT}/vector-store/{self.model.get_model_namespace()}"
        self._table_uri = f"{self._db_uri}/{self.model.get_model_name()}.lance"
        self._name = self.model.get_model_name()
        self._db = self._lance_connect()
        try: 
            self._table = LanceTable(self._db, self._name)
            if not fs.exists(self._table_uri):
                raise FileNotFoundError
        except FileNotFoundError as fex:
            if auto_register:
                logger.info(f"creating table {self._name}")
                self._table = self.table_from_schema(self._name, schema_or_model=self.model)
            else:
                raise
        
    def __repr__(self):
        return str(self._table)
            
    def table_from_schema(self, name: str, schema_or_model: pa.Schema | AbstractModel ):
        """
        given a pyarrow schema or abstract model, create the lance table
        """

        if not isinstance(schema_or_model, pa.Schema):
            schema_or_model = schema_or_model.to_arrow_schema()
        return self._db.create_table(name=name, schema=schema_or_model)
    
    def create_model(self, model: AbstractModel):
        return self.table_from_schema(model.get_model_name(), schema_or_model=model)

    def update_records(self, records: typing.List[AbstractModel], mode:str='append'):
        """
        adding tables to the lance store
        """
        if not records:
            return
        if not isinstance(records,list):
            records = [records]
            
        self._delete_by_record_key_values(records)
        
        """any fields we should excluded can be determined here"""
        excluded_fields = []
        def f(d):
            d = d.model_dump() if hasattr(d, "model_dump") else d
            return {k: v for k, v in d.items() if k not in excluded_fields}
            
        self._table.add(data=[f(r) for r in records], mode=mode)

    def select_one(self, id: str):
        pass

    def ask(self, question: str):
        return self.vector_search(question=question)
    
    def _delete(self, keys: typing.List[str]):
        """remove based on the model keys - assumed to be string like"""
        key_field = self.model.get_model_key_field() or 'name'
        in_list = ",".join([f'"{k}"' for k in keys])
        return self._table.delete(f"{key_field} IN ({in_list})")
        
        
    def _delete_by_record_key_values(self, records: typing.Optional[AbstractModel]):
        """get the key values in the list and delete match"""
        key_field = self.model.get_model_key_field() or 'name'
        keys = set(
                getattr(r, key_field) if hasattr(r, "dict") else r[key_field] for r in records
            )
        return self._delete(keys)
        
    def vector_search(
        self,
        question: str,
        search_operator: VectorSearchOperator = VectorSearchOperator.L2,
        limit: int = 7,
        probes: int = 20,
        refine_factor: int = 10,
        #distance threshold
    ):
        """
        lancedb search - see docs
        """
        
        """TODO: map all enumerations to the right thing for each provider """
        search_operator = 'l2'
        
        query_root = self._table.search(question).metric(search_operator)
        #todo hybrid
        #query_root.where(preds)
        query_root = (
            query_root.limit(limit).nprobes(probes).refine_factor(refine_factor)
        )
        """determine this from the model in general"""
        _extra_fields = ["name", "content"]
        
        if isinstance(_extra_fields, list) and _extra_fields:
            query_root = query_root.select(_extra_fields)

        """some work to do on the response model"""
        return query_root.to_list()
        


import os
from pathlib import Path
from importlib import import_module

POSTGRES_DB = "postgres"
POSTGRES_SERVER = "localhost"
POSTGRES_PORT = 5432
POSTGRES_PASSWORD = "password"
POSTGRES_USER = "sirsh"
POSTGRES_CONNECTION_STRING = f"postgresql://{POSTGRES_USER}:{POSTGRES_PASSWORD}@{POSTGRES_SERVER}:{POSTGRES_PORT}/{POSTGRES_DB}"
AGE_GRAPH = "funkybrain"
STORE_ROOT = os.environ.get('FUNKY_HOME',f"{Path.home()}/.funkyprompt")

def get_repo_root():
    """the root directory of the project"""
    path = os.environ.get("FUNKY_REPO_HOME")
    if not path:
        one = import_module("funkyprompt")
        if one.__file__ is not None:
            path = Path(one.__file__).parent.parent
        else:
            path = Path(__file__).parent.parent.parent
    return Path(path)


import requests
from bs4 import BeautifulSoup
from urllib.parse import urlparse
from funkyprompt.core.utils import logger

def _primary_image(soup):
    """a common format for finding a representative image"""
    og_image = soup.find('meta', property='og:image')
    if og_image and og_image.get('content'):
        return og_image['content']

    images = soup.find_all('img')
    if not images:
        return None

    largest_image = max(
        images,
        key=lambda img: int(img.get('width', 0)) * int(img.get('height', 0))
    )
 
    return largest_image.get('src')

def scrape_text(url):
    """
    simple text scraper - using this for primitive visit semantics
    test cases;
    
    
    """
    original_url = url
    def qualify(s, bridge='/'):
        """if images are not absolute, a lame attempt to make them so"""
        if original_url not in s:
            return f"{original_url.lstrip('/')}{bridge}{s.rstrip('/')}"
        
    url_parsed = urlparse(url)    
    if url_parsed.netloc == 'github.com':
        bridge = '/tree/main/'
        """a lame way to check for readme variants to replace the thing"""
        for r in ['README', 'Readme']:
            url_temp = f"{url_parsed.scheme}://raw.githubusercontent.com{url_parsed.path}/main/{r}.md"        
            response = requests.head(url, allow_redirects=True)
            if response.status_code == 200:
                url = url_temp
                logger.debug(f"Using url {url}")
                break
        
    response = requests.get(url)
    soup = BeautifulSoup(response.text, 'html.parser')
    return{
        'text' : soup.get_text(),
        'image': qualify(_primary_image(soup),bridge)
    }



"""
The runner can call the LLM in a loop and manage the stack of messages and functions
Function calling and streaming is handled
Open telemetry is used to publish metrics which a collector could manage
If this goes above 300 lines of codes we have failed - many of these lines are detailed comments.
"""

from funkyprompt.core import AbstractModel
from funkyprompt.services.models import language_model_client_from_context
from funkyprompt.core import utils
from funkyprompt.core.agents import (
    CallingContext,
    DefaultAgentCore,
    LanguageModel,
)
import json 
from funkyprompt.core import ConversationModel
from funkyprompt.services import entity_store
from . import MessageStack
from . import FunctionCall, FunctionManager, Function
import typing
import traceback

class Runner:
    """Runners are simple objects that provide the interface between types and language models
    The message setup is the only function that plays with natural language.
    While almost all of the "prompting" is pushed out to types and functions,
    This setup function is the one function you can play with to make sure the comms are right with the LLM.
    For example it is here we inject plans and questions and other hints for how to run things.
    But by design, the critical guidance should be abstracted by Types and Functions.
    Beyond this, the rest is routine;
    - import type metadata and functions from the model which controls most everything
    - run an executor loop sending context to the LLM
    - implement the invocation and message setup methods to manage the function and message stack
    Under the hood the function manager handles actual function loading and searching
    """

    def __init__(self, model: AbstractModel = None):
        """
        A model is passed in or the default is used
        The reason why this is passed in is to supply a minimal set of functions
        If the model has no functions simple Q&A can still be exchanged with LLMs.
        More general the model can provide a structured response format.
        This is powerful because the model is a Pydantic annotated type but can be realized as a json response
        """
        self.model = model or DefaultAgentCore()
        self._function_manager = FunctionManager()
        self.initialize()
        
    def __repr__(self):
        return f"Runner({(self.model.get_model_fullname())})"

    def initialize(self):
        """register the functions and other metadata from the model"""
        self._context = None
        """register the model's functions which can include function search"""
        self._function_manager.register(self.model)
        """the basic bootstrapping means asking for help, entities(types) or functions"""
        self._function_manager.add_function(self.lookup_entity)
        self._function_manager.add_function(self.help) # :)
        self._function_manager.add_function(self.activate_functions_by_name)
        """more complex things will happen from here when we traverse what comes back"""
    
    def activate_functions_by_name(self, function_name_to_entity_mapping: dict):
        """
        If you encounter a full name of a function that you don't have details for, you can activate it here.
        Once you activate it, it will be ready for use. Supply one or more function names to activate them.
        
        Args:
            function_name_to_entity_mapping (dict): provide a map between the entity name and the entity that the function belongs to
        """
        
        """NO-OP for now; the function manager can activate the functions - for now we are handling these cases when entities are loaded below.
           So this is just a placeholder if we want to have a more generalized registry and the idea of lazy activation to reduce cognitive load 
           - also, this function is a good way IF the agent thinks it has not got access to the function, give it something to do and then nudge it with the message below.
        """
        fm = function_name_to_entity_mapping
        if not isinstance(fm,dict):
            raise Exception("When calling this function, in `function_names` you must provide a mapping between the function and the entity it belongs e.g. {'some_function': 'namespace.entity'}")
        
        utils.logger.debug(f'activating function {fm}')
        fm = self._function_manager.add_functions_by_name(fm)
            
        return {
            'status': f"Re: the functions {list(fm.values())}, now ready for use. please go ahead and invoke."
        }
        
    def lookup_entity(self, name:str):
        """lookup entity by one or more keys
        
        Args:
            name: one or more names to use to lookup the entity or entities 
        """
        
        from funkyprompt.services import entity_store
       
        """todo test different parameter inputs e.g. comma separated"""
        entities =  entity_store(self.model).get_nodes_by_name(name)
        
        """register entity functions if needed and wait for the agent to ask to activate them"""
        for e in entities:
            self._function_manager.register(e)
        
        """when we return the entities, its better to return them with metadata (as opposed to just fetching the record data only)"""
        return AbstractModel.describe_models(entities)

    def help(self, questions: str | typing.List[str]):
        """if you are stuck ask for help with very detailed questions to help the planner find resources for you.

        Args:
            question (str): provide detailed questions to guide the planner
        """

        utils.logger.debug(f"help/{questions=}")

        try:
            """for now strict planning is off"""
            plan = self._function_manager.plan(questions,strict=False)
        except:
            utils.logger.warning(f"failing to plan - {traceback.format_exc()}")
            return {"message": "planning pending - i suggest you use world knowledge"}

        """describe the plan context e.g. its a plan but you need to request the functions and do the thing -> update message stack"""

        return plan

    def invoke(self, function_call: FunctionCall):
        """Invoke function(s) and parse results into messages

        Args:
            function_call (FunctionCall): the payload send from an LLM to call a function
        """
        f = self._function_manager[function_call.name]
        
        if not f:
            message = f"attempting to load function {function_call.name} which is not activated - please activate it"
            utils.logger.warning(message)
            data = MessageStack.format_function_response_error(
                function_call.name, ValueError(message), self._context
            )
        else:

            try:
                """try call the function - assumes its some sort of json thing that comes back"""
                data = f(**function_call.arguments) or {}
                data = MessageStack.format_function_response_data(
                    function_call.name, data, self._context
                )
                """if there is an error, how you format the message matters - some generic ones are added
                its important to make sure the format coincides with the language model being used in context
                """
            except TypeError as tex: #type errors are usually the agents fault
                utils.logger.warning(f"Error calling function {tex}")
                data = MessageStack.format_function_response_type_error(
                    function_call.name, tex, self._context
                )
            except Exception as ex: #general errors are usually our fault
                utils.logger.warning(f"Error calling function {traceback.format_exc()}")
                data = MessageStack.format_function_response_error(
                    function_call.name, ex, self._context
                )

        """update messages with data if we can or add error messages to notify the language model"""
        self.messages.add(data)

    @property
    def functions(self) -> typing.Dict[str, Function]:
        """provide access to the function manager's functions"""
        return self._function_manager.functions

    def run(self, question: str, context: CallingContext, limit: int = None):
        """Ask a question to kick of the agent loop"""

        """setup all the bits before running the loop"""
        lm_client: LanguageModel = language_model_client_from_context(context)
        self._context = context
        self.messages = MessageStack(
            model=self.model,
            question=question,
            current_date=utils.dates.now(),
            function_names=self.functions.keys(),
            language_model_provider=context.model,
        )

        """run the agent loop to completion"""
        for _ in range(limit or context.max_iterations):
            response = None
            function_descriptions = [f.to_json_spec() for f in self.functions.values()]
            """call the model with messages and function + our system context"""
            response = lm_client(
                messages=self.messages.model_dump(),
                context=context,
                functions=function_descriptions,
            )
            if isinstance(response, FunctionCall):
                """call one or more functions and update messages"""
                self.invoke(response)
                continue
            if response is not None:
                # marks the fact that we have unfinished business
                break

        """fire telemetry"""

        """log questions to store unless disabled"""
        self.dump(question, response, context)
        
        """call the observer - often a no-op"""
        response = self.observe_structure(response)

        return response
    
    def observe_structure(self, response):
        """
        Particularly when the response is structured, if the user is not a system user, we may want to call the entity explainer
        - response = self.model.explain(response, explain_json_only=true)
        
        We may want to go the other way to when answers are unstructured to structure (and save) but that is probably already built in to the type behaviour
        - we describe types as "extractive" in that we can take unstructured data and structure and save it as a core type capability
        
        structured <--> unstructured decision making
        """
        return response

    def dump(self, questions: str, response: str, context: CallingContext):
        """dumps the messages and context to stores
        if the session is a typed objective this is updated in a slowly changing dimension
        generally audit all transactions unless disabled
        """
        from uuid import uuid4

        default_id = str(uuid4())

        """for any pydantic response"""
        if hasattr(response, "model_dump_json"):
            """dumpy state"""
            response = response.model_dump_json()

        entity_store(ConversationModel).update_records(
            ConversationModel(
                id=default_id,
                user_id=context.username or "system",
                objective_node_id=context.session_id,
                content={"question": questions, "response": response},
            )
        )


    def __call__(
        self, question: str, context: CallingContext = None, limit: int = None
    ):
        """
        Ask a question to kick of the agent loop
        """
        context = context or CallingContext()
        return self.run(question, context, limit=limit)

    def explain(cls, data:dict|typing.List[dict]):
        """
        it is often convenient to run the runner directly on data which uses this standard recipe
        """
        P = f"please explain the following data according to your guidelines - ```{json.dumps(data)}``` and respond in a json format - use the model provided"
        return cls(P)
        
"""
#TODO: general test e.g. call a function with limit=1, it doesnt loop and confirm the result and therefore we might not get behaviour expected
generally we want to ensure consistent output    
the response should always be JSON but revert to a simple form {response=}
"""



"""
A simpler s3 and file system interface. read, write, ls, copy and basic file system operations are supported
for example we dont have upload or download since the copy format is used instead e..g copy rom file:/// to s3:///
we dont have zip or unzip since the format of copying to an archive is supported

"""

from funkyprompt.services.fs.s3 import S3Provider
import polars as pl
import pyarrow.dataset as ds
from PIL import Image
from pathlib import Path
import io
import typing
import requests


def is_s3(uri):
    return uri[:5] == "s3://"


def is_archive(uri):
    return ".zip" in uri.lower() or ".tar" in uri.lower() or ".gz" in uri.lower()


class FS:
    """
    entry point to file systems to abstract away from local or S3 - all methods may not yet be implemented as we normally use S3
    """

    def https_to_file(self, web_request_uri, target_uri, token=None):
        """
        if we have a remote resource we can write it to a file in our file system e.g.
        - download airtable or slack files to s3
        """
        response = (
            requests.get(web_request_uri)
            if not token
            else requests.get(
                web_request_uri, headers={"Authorization": "Bearer %s" % token}
            )
        )

        with self.open(target_uri, "wb") as f:
            f.write(response.content)

    def open(self, uri: str, mode: str):
        """
        open files for read or write on the filesystem
        """
        if is_s3(uri):
            return S3Provider().open(uri, mode=mode)
        else:
            return open(uri, mode)

    def exists(self, uri):
        """
        check if file or folder exists on the file system
        """
        if is_s3(uri):
            return S3Provider().exists(uri)

        return Path(uri).exists()

    def read(self, uri: str, **options):
        """
        read any data type - extensions are used to determine the reader
        """

        if is_s3(uri):
            return S3Provider().read(uri, **options)

        from cairosvg import svg2png

        P = Path(uri)
        with open(uri, mode=options.get("mode", "r")) as f:
            if P.suffix in [".svg"]:
                return Image.open(io.BytesIO(svg2png(f.read(), **options)))

        # check extensions - svg example

    def write(self, uri, data, **options):
        """
        write any data type - extensions are used to determine the writer
        """
        if is_s3(uri):
            return S3Provider().write(uri, data, **options)

    def copy(self, uri_from: str, uri_to: str):
        """
        overloaded copy between file systems and locations
        """
        if is_s3(uri_from) and is_s3(uri_to):
            return S3Provider().copy(uri_from, uri_to)
        if not is_s3(uri_from) and not is_s3(uri_to):
            raise Exception(f"Not handled mode {uri_from} -> {uri_to}")
        if is_s3(uri_from) and not is_s3(uri_to):
            # s3 handles download
            return S3Provider().copy(uri_from, uri_to)
        if not is_s3(uri_from) and is_s3(uri_to):
            # s3 handles upload
            return S3Provider().copy(uri_from, uri_to)

        raise Exception(f"Not handled mode {uri_from} -> {uri_to}")

    def cache_data(self, data: typing.Any, **kwargs):
        """
        cache data is assumed to be on s3
        """
        return S3Provider().cache_data(data, **kwargs)

    def ls(self, uri: str, **options):
        """
        list files from a prefix `uri`. To look at immediate children use `ls_dirs`
        """
        if is_s3(uri):
            return S3Provider().ls(uri, **options)

    def ls_dirs(self, uri: str, **options):
        """
        list the directories that are child to the prefix `uri`
        """
        if is_s3(uri):
            return S3Provider().ls_dirs(uri, **options)

    def ls_iter(self, uri: str, **options):
        """
        for paginated use cases (TODO:)
        """
        if is_s3(uri):
            for item in S3Provider().ls_iter(uri, **options):
                yield item

    def delete(self, uri: str, limit=100):
        """
        delete objects in a folder/dir - a safety `limit` is used
        this can be used to delete files or directory contents
        """
        if is_s3(uri):
            return S3Provider().delete(uri, limit=limit)

    def read_dataset(self, uri) -> ds.Dataset:
        """
        read data such as parquet files as a pyarrow dataset
        """
        if is_s3(uri):
            return S3Provider().read_dataset(uri)

    def read_image(self, uri) -> Image:
        """
        read image as a PIL image
        """
        if is_s3(uri):
            return S3Provider().read_image(uri)

        return Image.open(uri)

    def apply(self, uri: str, fn: typing.Callable):
        """apply an op to a file like object"""
        if is_s3(uri):
            return S3Provider().apply(uri, fn)

    def local_file(self, uri):
        if is_s3(uri):
            filename = Path(uri).name
            filename = f"/tmp/{filename}"
            S3Provider().copy(uri, filename)
            return filename
        return uri



from pydantic import BaseModel, create_model, Field, model_validator
import uuid
import typing
from funkyprompt.core.types import inspection
from funkyprompt.core.types.sql import SqlHelper
from funkyprompt.core.types.cypher import CypherHelper
from funkyprompt.core.utils.ids import funky_id
from funkyprompt.core.utils.dates import utc_now
import datetime
from pydantic._internal._model_construction import ModelMetaclass
import json
"""
names are always unique in funkyprompt for key-value entity lookups
however there are times when the name is not unique and if so the id should be generated by the system
"""
KEY_FIELD_ATTRIBUTE = "is_key"
DEFAULT_KEY_ATTRIBUTE_NAME = "name"
DEFAULT_NAMESPACE = "default"

def can_parse_json(d):
    """dumb validator"""
    try:
        json.loads(d)
        return True
    except:
        return False
    
class Node(BaseModel):
    """a simple node"""
    name: str
    node_type: str = Field(default='generic')
    attributes: typing.Optional[str|dict] = None
    
    @property
    def key(self):
        return f"{self.node_type}_{self.name}"
    
    @model_validator(mode="before")
    @classmethod
    def _types(cls, values):
        """formatting for cypher and often we have qualified node names and this makes it easier"""
        values['node_type'] = values.get('node_type', 'generic').replace('.','_')
        att = values.get('attributes')
        if isinstance(att,dict):
            values['attributes'] = Node.format_attributes(att)
        return values

    @staticmethod
    def format_attributes(att):
        def f(s):
            return s if not isinstance(s,str) else f'"{s}"'
        return " ".join([f"k:{f(v)}" for k,v in att.items()]) if att else None
    
class Edge(BaseModel):
    """a simple edge"""
    source_node: Node
    target_node: Node
    attributes: typing.Optional[str|dict] = Field(default=None)
    description: str
    type: typing.Optional[str] = Field(default='edge')
    
    @property
    def edge_name(self):
        """when edges are created they are typed - we can maintain an one edge of type between two unique nodes but we could qualify"""
        return f"{self.source_node.key}{self.target_node.key}"

    @staticmethod
    def format_attributes_for_assignment(att, alias='e'):
        def f(s):
            return s if not isinstance(s,str) else f'"{s}"'
        return " ".join([f"{alias}.{k}={f(v)}" for k,v in att.items()]) if att else None
    
    def make_edge_upsert_query_fragment(self: "Edge", index:int=0):
        """Conventional generation of edges uses a form of cypher to format,
        the node name is assumed to be the key in funkyprompt,
        attributes are upserted on the name of the edge
        """

        """TODO: it may ne we want to understand edge uniqueness but for now we create liberally 
        - we assume the edge name can be generated uniquely but attributes can be updated
        """
        attribute_assignments = Edge.format_attributes_for_assignment(self.attributes, alias=f"e{index}") or f"""e{index}.description="{self.description.replace('"', '/"')}" """
        attribute_assignments += f",e{index}.timestamp='{utc_now().isoformat()}'"
        A = f"""a{index}:{self.source_node.node_type} {{name: '{self.source_node.name}' }}"""
        B = f"""b{index}:{self.target_node.node_type} {{name: '{self.target_node.name}' }}"""
        return f"""MERGE ({A})-[e{index}:{self.type} {{name: '{self.edge_name}' }}]->({B}) SET {attribute_assignments}"""
        
class AbstractModel(BaseModel):
    """"""
    
    """should have a name for the naming logic to work and still have config take precedence"""
    # class Config:
    #     name: str = "abstract_model"
    #     namespace: str = "core"
    #     description: str = "Provide a rich model description"

    id: typing.Optional[str | uuid.UUID] = Field(
        default=None,
        description="A unique hash/uuid for the entity. The name can be hashed if its marked as the key",
    )

    @classmethod
    def get_model_name(cls):
        c = getattr(cls, "Config", None)
        if c and getattr(c, "name", None):
            return c.name
        """else infer from lib"""
        s = cls.model_json_schema(by_alias=False)
        return s.get("title", cls.__name__)

    @classmethod
    def get_model_namespace(cls):
        c = getattr(cls, "Config", None)
        if c and getattr(c, "namespace", None):
            return c.namespace
        """else infer from lib"""
        # convention
        namespace = cls.__module__.split(".")[-1]
        return (
            namespace
            if namespace not in ["model", "__main__", "entity"]
            else DEFAULT_NAMESPACE
        )

    @classmethod
    def get_model_description(cls):
        """the description of the entity - import for prompting"""
        c = getattr(cls, "Config", None)
        if c and getattr(c, "description", None):
            return c.description

    @classmethod
    def get_model_fullname(cls):
        """
        the model name is our convention e.g. meta.bodies
        usually we determine this from either a config object or the type itself
        """
        return f"{cls.get_model_namespace()}.{cls.get_model_name()}"

    @classmethod
    def get_type_fullname(cls):
        """
        the object name
        """
        return f"{cls.get_model_namespace()}.{cls.get_model_name()}"

    @classmethod
    def get_model_key_field(cls):
        """
        the field that is used as the primary key if it exists
        - otherwise a unique id should be generated by the system
        all models have ids and some of additional friendly names
        """
        s = cls.model_json_schema(by_alias=False)
        key_props = [
            k for k, v in s["properties"].items() if v.get(KEY_FIELD_ATTRIBUTE)
        ]
        if len(key_props):
            return key_props[0]

    def get_key_value(cls):
        """
        return the instance key value based on the configured key attribute name
        """
        return getattr(cls, cls.get_model_key_field())

    @classmethod
    def create_model(cls, name: str, namespace: str = None, **fields):
        """
        For dynamic creation of models for the type systems
        create something that inherits from the class and add any extra fields
        """
        namespace = namespace or cls.get_model_namespace()
        return create_model(name, **fields, __module__=namespace, __base__=cls)

    # def get_dynamic_functions

    def get_dummy_values(cls):
        """dummy values useful in some automation"""
        pass

    def to_arrow(cls):
        """convert the object to its pyarrow representation"""
        pass

    @classmethod
    def sql(cls) -> SqlHelper:
        """reference the sql helper"""

        return SqlHelper(cls)

    @classmethod
    def cypher(cls) -> CypherHelper:
        """reference the cypher helper"""

        return CypherHelper(cls)

    def db_dump(self):
        """serialize complex types as we need for DBs/Postgres
        - we do things like allow for config to turn fields off
        - we map complex types to json
        - embedding are added async on a new table in our model

        """
        from funkyprompt.core.types.sql import SqlHelper
        import json

        data = vars(self)
        """control selectable fields by exclude or other attributes"""
        fields = SqlHelper.select_fields(self)

        def check_complex(v):
            if isinstance(v, dict):# or isinstance(v, list):
                return json.dumps(v)
            return v

        data = {k: check_complex(v) for k, v in data.items() if k in fields}

        return data
    
    @classmethod
    def to_arrow_schema(cls):
        """
        get the arrow schema from the pydantic type
        """
        from funkyprompt.core.types.pydantic import pydantic_to_arrow_schema

        return pydantic_to_arrow_schema(
            cls
        )
        
    @classmethod
    def get_embedding_fields(cls) -> typing.Dict[str, str]:
        """returns the fields that have embeddings based on the attribute - uses our convention"""
        needs_embeddings = {}
        for k, v in cls.model_fields.items():
            extras = getattr(v, "json_schema_extra", {}) or {}
            if extras.get("embedding_provider"):
                needs_embeddings[k] = f"{k}_embedding"
        return needs_embeddings

    @classmethod
    def _get_child_models(cls) -> typing.List["AbstractModel"]:
        """
        if this is implemented, we will show the child types in the model description
        """
        return []
    
    @classmethod
    def get_model_as_prompt(cls) -> str:
        """the model as prompt provides a schema and also the description of the model
        if the base class implements `_get_prompting_data` then data will be loaded into context
        For example this is used in function planning

        this is experimental and may not be a perfect abstraction
        """
        
        from funkyprompt.core.types.pydantic import get_pydantic_properties_string

        injected_data = ""
        if getattr(cls, "_get_prompting_data", None) is not None:
            injected_data = cls._get_prompting_data()

        return f"""## About the model: {cls.get_model_name().upper()}
    
### Description

{cls.get_model_description()}
    
### About the types and fields you in the response model (JSON)

_the child types will appear first followed by the parent model to use_

{get_pydantic_properties_string(cls, cls._get_child_models())}

{injected_data}
    """

    #############
    ##   INSPECTION
    #############

    @classmethod
    def get_class_and_instance_methods(cls):
        """returns the methods on the type that we care about"""

        # TODO: by default we do not show all base methods as agent-callable but we can register run_search and add here (crud)

        if not isinstance(cls, ModelMetaclass):
            cls = cls.__class__
            
        """any methods that are not on the abstract model are fair game"""
        methods = inspection.get_class_and_instance_methods(cls, inheriting_from=AbstractModel)

        """return everything but hide privates"""
        return [m for m in methods if not m.__name__[:1] == "_"]

    """
    ----------
    """

    @classmethod
    def _register(cls):
        """a not to be abused but convenient self-register in the core entity store"""
        from funkyprompt.services import entity_store

        return entity_store(cls)._create_model()
    
    @classmethod
    def _ask(cls, question:str, raw_results:bool=False):
        """convenience method to load up an entity store with the model and 
        ask a question and optionally run the wrong store query
        this is hidden for now so as not to harden this interface
        """
        from funkyprompt.services.models import language_model_client_from_context
        from funkyprompt.services import entity_store
        from funkyprompt.core.agents import   MessageStack, LanguageModel
        result = entity_store(cls).ask(question) 
        if raw_results:
            return result
        messages= MessageStack.from_q_and_a(question, result)
        lm_client: LanguageModel = language_model_client_from_context(None)
        response = lm_client(messages=messages, functions=None, context=None)
        return response 
    
    @classmethod
    def _describe_model(cls, qualify_function_name: bool = False):
        """
        This model takes a particular format for allow an agent to know what it can do with the model
        it could convert to get-prompt but we will see
        """
        #
        """now prepare the metadata starting with functions"""
        functions = []
        for f in cls.get_class_and_instance_methods():
            name = f.__name__
            if qualify_function_name:
                name = f"{cls.get_model_namespace()}_{cls.get_model_name()}_{f.__name__}"
            functions.append({
                'function_name': name ,
                'description' : f.__doc__,
                'entity_name': cls.get_model_fullname()
            })
        
        """add the models into a map, unique on the entity type"""
        return {
            'about': cls.get_model_description(),
            'functions': functions,
            'fields': {k: v.description for k,v in cls.model_fields.items()}
        }
    
    @classmethod
    def describe_models(cls, models : typing.List["AbstractModel"],**kwargs)->dict:
        """
        Given a collection of models, we return the model data collection but also provide the distinct list of entity metadata
        """        
        
        if not models:
            return {
            'type_metadata': {},
            'status': 'no data matched the query',
            'records': []
        }
        
        if not isinstance(models,list):
            models = [models]
        
        metadata = {}
        data = []
        for m in models:
            """dump and qualify the data"""
            d = m.model_dump()
            d['model_type'] = m.get_model_fullname()
            data.append(d)
            """add the models into a map, unique on the entity type"""
            metadata[m.get_model_fullname()] = m._describe_model()
        return {
            'type_metadata': metadata,
            'records': data
        }
        
    @classmethod
    def explain(cls, data:str|dict|typing.List[dict], explain_json_only:bool=False, **kwargs)->str:
        """
        it is often convenient to run the runner directly on data which uses this standard recipe below;
        - the model is used to provided context
        - the data are provided and can be explained
        
        this is a useful pattern to de-structure internal agent communication for the user
        
        Args:
            data: the data is anything that can be dumped as json or a raw string
            explain_json_only: (default - False) to only attempt to explain structured data and "trust" simple text as being explanatory
        """
        from funkyprompt.services.models import language_model_client_from_context
        from funkyprompt.core.agents import LanguageModel
        
        if explain_json_only and not can_parse_json(data):
            return data
        
        P = f"""Please explain the data below according to the `Guidelines` provided at bottom...
        
        ## Data
        
         ```{data if isinstance(data,str) else json.dumps(data)}```
         
        ## Guidelines
        
        {cls.describe_models(data)}
        """

        lm_client: LanguageModel = language_model_client_from_context(None)
        """an example where messages can be any simple textual or json thing"""
        response = lm_client(messages=P, functions=None, context=None)
        return response 

class AbstractEntity(AbstractModel):
    """the abstract entity is a sub class of model that admits a unique name0
    - abstract entities are treated as graph nodes unless the config specifies exclude_from_graph=True

    """

    name: str = Field(description="The name is unique for the entity", is_key=True)

    @classmethod
    def _lookup_entity(cls, name:str, include_relations: bool=False):
        """convenience to lookup the type by name (cypher query on the node)"""
        from funkyprompt.services import entity_store
        ntype = f"{cls.get_model_namespace()}_{cls.get_model_name()}"
        query = f"""MATCH (v:{ntype} {{name:'{name}'}}) RETURN v"""
        if include_relations:
            query = f"""MATCH (v:{ntype} {{name:'{name}'}})-[r]-(o) RETURN v,r,o"""
            return entity_store(cls).query_graph(query,returns=['n', 'r', 'o'])
        return entity_store(cls).query_graph(query)
    
    @model_validator(mode="before")
    @classmethod
    def _id(cls, values):
        """"""
        if not values.get("id"):
            """very important to observe current convention of 
            id generated from caseless string"""
            values["id"] = funky_id(values["name"].lower())
        return values

    @classmethod
    def run_search(
        cls, questions: str | typing.List[str], limit: int = None, **kwargs
    ) -> typing.List[AbstractModel]:
        """search the entity using the default store

        Args:
            questions (str | typing.List[str]): ask one or more questions - the more the better
            limit (int, optional): provide an optional search limit. Defaults to None.
        """

        from funkyprompt.services import entity_store

        return entity_store(cls).ask(questions, limit, **kwargs)

    @classmethod
    def upsert_entity(cls, name:str, data_delta: dict)->"AbstractModel":
        """Save the entity by merging new and old data to the final object.
        You should always lookup the old entity if you do not already have it OR you should check the schema of the object to save a single dictionary object!!
        
        You should be efficient by sending only the changed fields.
        If a field is not changed omit it. 
        If a field has extra content, show the combined content for the field.
        
        Args:
            name: str : the unique name of the object
            data_delta (dict): the object delta

        """
        from funkyprompt.services import entity_store
        
        """the rationale here is not so much to save, 
           which we can do ourselves, but merge the old and new context from the language model
           another option would be to send deltas which we could load and merge the existing
           the name is added above to allow new objects to be saved and to force the inclusion of the name in the changeset
           we can recover conversations about the object by id
           on the one we want to keep it simple and fuzzy but we dont want to drop important facts
           BIG QUESTION: Ability to merge memory!
           """
        store = entity_store(cls)
      
        existing = store.select_one(name) 
       
        if existing:
            existing = existing.model_dump()
            existing.update(data_delta)
            existing['name'] = name
        else:
            data_delta['name'] = name
            try:
                """this validation is primarily to make sure that the caller truly understand the schema """
                existing = cls(**data_delta).model_dump()
            except:
                raise ValueError(f"You have supplied a value {data_delta} that is not compatible with the type schema for {cls.get_model_fullname()}")
       
        response =  store.update_records(cls(**existing))
        if response:
            """assumed contract on update one"""
            response = response[0]
        """if the response is failing there may be times you want to see 
        what the attributes are for now we are strict"""
        
        response = cls.model_validate(response)
      
        return {
            'status': 'the entity has been updated',
            'value': response
        }
    

class AbstractContentModel(AbstractEntity):
    """use to generate generic content types
    This is useful because we can create generic types to save in the data stores in their own tables
    These have content blobs that can be saved with the open ai embedding

    Example

        ```python
        my_generic =  AbstractContentModel.create_model(name='test', namespace='public')
        #creates a new table public.test with the content model schema
        my_generic._register()
        # you can now create these objects and insert them
        ```
    """

    content: str = Field(
        description="The name is unique for the entity", embedding_provider="openai"
    )
    category: typing.Optional[str] = Field(
        default=None, description="the grouping category for the content"
    )


class AbstractImageContentModel(AbstractContentModel):
    """like the abstract content model but for image data"""

    content: str = Field(description="Image uri", embedding_provider="clip")
    description: typing.Optional[str] = Field(
        default=None,
        description="optional image description",  # TODO: may add an open embedding to this
    )


class AbstractEdge(BaseModel):
    
    timestamp: datetime.datetime
    source_name: str
    target_name: str



"""implement some basic batch merge statements - they generally take the conventional forms

MERGE (a:Type{name:x})-[e:Type({name:y}]-(b:Type{name:y})], SET a.property = 'value'
where each variable a,e,b is indexed in the batch e.g. a0, a1, a2...

funkyprompt uses a very simple type of graph since its hybrid
we are simply registering keys and relationships between them 
so we dont write a lot of properties in the graph except for system props

The type system works with the AbstractEntity to qualify node and edge types

This helps agents find things it would otherwise lose

"""

import typing

class CypherHelper:
    """
    see [age docs](https://age.apache.org/age-manual/master/intro/overview.html)
    """

    def __init__(self, model, db=None):
        """"""
        from funkyprompt.core import AbstractEntity

        self.model: AbstractEntity = model

    def query_from_natural_language(cls, question: str):
        """"""
        return None

    def get_graph_model_attribute_setter(self, node, has_full_entity:bool=False, alias='n'):
        """the default node behviour is to just keep the name but we can 'index' other attributes
        this can either be done upfront or later on some trigger or job
        """
        #this is a marker that shows we have associated a full entity with the node
        attributes = f"{alias}.entity = 1" if has_full_entity else None
        
        if attributes:
            return f"""SET {attributes}"""
        return ''

    # def upsert_path_query(self, node):https://age.apache.org/age-manual/master/clauses/create.html

    def create_script(self):
        """create the node - may well be a no-op but we register anyway"""
        label = self.model.get_model_fullname().replace(".", "_")
        q = f"""
        """
        return None
    
    def upsert_relationship_query(self, relationships):
        """
        the entities may contain relationships or may not
        """
        return None
    
    def upsert_nodes_statement(self, nodes, label:str=None, has_full_entity:bool=False):
        """
        create a node upsert query - any attributes can be upserted
        but labeled nodes are supposed to be unique by name
        
        Args:
            nodes: a list of entities 
            label: its assumed the label is based on the model but can be anything
            has_full_entity: a tracker to see if we are also adding the entity or just making a relationship to a node 
        """

        if not isinstance(nodes, list):
            nodes = [nodes]

        label = label or self.model.get_model_fullname().replace(".", "_")
        
        cypher_queries = []

        i = 0
        for i,n in enumerate(nodes):
            """any nodes can be added for any time but often we do a batch based on the source model
            see Edge and Node types to understand the interface for getting node_type
            """
            applied_label = getattr(n, 'node_type', label).replace('.','_')
            """we may set some attributes like descriptions and stuff"""
            cypher_queries.append(
                f"""MERGE (n{i}:{applied_label} {{name: '{n.name}'}})
                { self.get_graph_model_attribute_setter(n, alias=f"n{i}", has_full_entity=has_full_entity)  }
               """
            )

        """for now block them up but we investigate a more efficient notation
           we generally expect low cardinality upserts for entity types in practice
        """
        if cypher_queries:
            """fetch a sample"""
            return "\n".join(cypher_queries) + f" RETURN n{i}"

    def upsert_edges_statement(self, edges):
        """
        create a batch upsert that returns the last one sample
        """
        if edges:
           return "\n".join([e.make_edge_upsert_query_fragment(index=i) for i, e in enumerate(edges)]) + f" RETURN e{len(edges)-1}"
        
    def distinct_edges(self, entities: typing.List[typing.Any]):
        """
        uses the model annotation conventions to extract nodes related to item.
        fields should be annotated with `RelationshipField` 
        and then we can either have anonymous edges as dicts or lists of AbstractEntity
        """
        from funkyprompt.core.AbstractModel import Node, Edge
        from funkyprompt.core import AbstractEntity
        from funkyprompt.core.fields.annotations import AnnotationHelper
  
        if not entities:
            return []
              
        if not isinstance(entities, list):
            entities: typing.List[AbstractEntity] = [entities]
        items = []
        for e in entities:
            a = AnnotationHelper(e)
            for field_name, edge_type in a.typed_edges.items():
                edge_type = edge_type or f"HAS_{field_name.upper()}" #f"HAS_{t.get_model_name().upper()}"
                """assume relationships are dicts for now"""
                field = getattr(e, field_name, None) or {}
                """if the field type is a dict 
                - otherwise we extract nodes and edges in a different way"""
                if isinstance(field, dict):
                    for target_name, description in field.items():
                        items.append(
                            Edge(source_node=Node(name=e.name,node_type=e.get_model_fullname() ),
                                #we dont know the target node type yet
                                target_node=Node(name=target_name),
                                #this is the edge description
                                description=description,
                                type=edge_type))
                #the case that is based on the typed child entities that have relations - strings are treated like generic backlinks or resources
                elif isinstance(field, str) or isinstance(field, list) or isinstance(field, AbstractEntity):
                    if not isinstance(field,list):
                        field: typing.List[AbstractEntity] = [field]
                    """now we have a collection of abstract 'child' entities that can be added as typed relationships
                       These will have the form HAS_TYPE -> TARGET_TYPE(name)
                    """
                    for t in field:
                        if isinstance(e, AbstractEntity):
                            items.append(
                                Edge(source_node=Node(name=e.name,node_type=e.get_model_fullname() ),
                                     target_node=Node(name=t.name, node_type=t.get_model_fullname()),
                                     type=edge_type))
                        #string table are just labels or back links
                        if isinstance(e,str):
                            items.append(
                                Edge(source_node=Node(name=e,node_type='public.resource' ),
                                     target_node=Node(name=t.name, node_type=t.get_model_fullname()),
                                     type=edge_type))
                #a third type would be a tuple instead of a dict to describe the agent - we could add this thing to the runner
        return items

    def upsert_relationships_queries(self, entities):
        """
        relationships register nodes that do not exist and add relationships between the source and target nodes
        this is how we create graph bottoms and its ok for them to be not fully connected
        """
        
        from funkyprompt.core.AbstractModel import Edge
            
        """extract the nodes and the edges from the entities"""
        
        edges: typing.List[Edge] = self.distinct_edges(entities)
        """get distinct nodes used in the relationships"""
        new_nodes = list({e.target_node.key : e.target_node for e in edges}.values())        
        edges_query = self.upsert_edges_statement(edges)
        node_query = self.upsert_nodes_statement(new_nodes)
        
        """return both queries for execution"""
        return node_query,edges_query
        


       
        


from pydantic import BaseModel, Field
import typing
import datetime
from funkyprompt.core import AbstractEdge

class ProjectTask(AbstractEdge):
    class Config:
        name: str = 'project_task'
        namespace: str = 'public'
        description: str = "When a user associates a task with a project or goal"
        
    """edges normally are weighted"""
    utility: typing.Optional[float] = Field(default=None, description="If the utility of the task can be estimated for the user's project or goals")
    effort: typing.Optional[float] = Field(default=None, description="An estimate of the difficulty of the task given what has been done so far")
   
class TaskResource(AbstractEdge):
    class Config:
        name: str = 'task_resource'
        namespace: str = 'public'
        description: str = "When a user associates a resource with a task"
        
    """edges normally are weighted"""
    utility: typing.Optional[float] = Field(default=None, description="If the utility of the resource can be estimated for the user's task")
   

class ResourcePerson(AbstractEdge):
    class Config:
        name: str = 'resource_person'
        namespace: str = 'public'
        description: str = "Associate a person with a resource for example a thought leader or owner"
    
    #we could sub class to create categories but the association is fine for now
    """edges normally are weighted"""
    reputation: typing.Optional[float] = Field(default=None, description="Teh reputation or rank of the person with respect to the resource")
    category: typing.Optional[str] = Field(default=None,description="How the person is associated with the idea - THINKER|WORKER|OWNER")
     
     



import typing
import pydantic
import pyarrow as pa
from datetime import date, datetime
import sys
import inspect
import types

def get_pydantic_properties_string(cls, child_types=None):
    """
    this is useful as a prompting device
    """
    annotations = typing.get_type_hints(cls)
    
    """if known child types are provided, we render them first"""
    child_strings = f"\n\n".join(get_pydantic_properties_string(t) for t in child_types or [])
    
    class_str = f"\n\nclass {cls.__name__}(BaseModel)\n"
    for field_name, field_type in annotations.items():
        field_default = getattr(cls, field_name, ...)
        field_info = cls.__fields__.get(field_name)
        description = (
            f" # {field_info.description}"
            if getattr(field_info, "description", None)
            else ""
        )
        type_str = repr(field_type)

        if field_default is ...:
            class_str += f"  -  {field_name}: {type_str}{description}\n"
        else:
            if isinstance(field_default, pydantic.Field):

                class_str += f" - {field_name}: {type_str} = Field(default={repr(field_default.default)}) {description}\n"
            else:
                class_str += f" - {field_name}: {type_str} = {repr(field_default)} {description}\n"
    return child_strings + class_str


def get_extras(field_info, key: str):
    """
    Get the extra metadata from a Pydantic FieldInfo.
    """
    return (field_info.json_schema_extra or {}).get(key)


def _py_type_to_arrow_type(py_type, field, coerce_str=True):
    """Convert a field with native Python type to Arrow data type.

    Raises
    ------
    TypeError
        If the type is not supported.
    """
    if py_type == int:
        return pa.int64()
    elif py_type == float:
        return pa.float64()
    elif py_type == str:
        return pa.utf8()
    elif py_type == bool:
        return pa.bool_()
    elif py_type == bytes:
        return pa.binary()
    elif py_type == date:
        return pa.date32()
    elif py_type == datetime:
        tz = get_extras(field, "tz")
        return pa.timestamp("us", tz=tz)
    elif getattr(py_type, "__origin__", None) in (list, tuple):
        child = py_type.__args__[0]
        return pa.list_(_py_type_to_arrow_type(child, field))

    if coerce_str:
        return pa.utf8()

    raise TypeError(
        f"Converting Pydantic type to Arrow Type: unsupported type {py_type}."
    )


def is_nullable(field) -> bool:
    """Check if a Pydantic FieldInfo is nullable."""
    if isinstance(field.annotation, typing._GenericAlias):
        origin = field.annotation.__origin__
        args = field.annotation.__args__
        if origin == typing.Union:
            if len(args) == 2 and args[1] == type(None):
                return True
    elif sys.version_info >= (3, 10) and isinstance(field.annotation, types.UnionType):
        args = field.annotation.__args__
        for typ in args:
            if typ == type(None):
                return True
    return False


def _pydantic_model_to_fields(model: pydantic.BaseModel) -> typing.List[pa.Field]:
    return [_pydantic_to_field(name, field) for name, field in model.__fields__.items()]


def _pydantic_to_arrow_type(field) -> pa.DataType:
    """Convert a Pydantic FieldInfo to Arrow DataType"""

    if isinstance(field.annotation, typing._GenericAlias) or (
        sys.version_info > (3, 9) and isinstance(field.annotation, types.GenericAlias)
    ):
        origin = field.annotation.__origin__
        args = field.annotation.__args__
        if origin == list:
            child = args[0]
            return pa.list_(_py_type_to_arrow_type(child, field))
        elif origin == typing.Union:
            if len(args) == 2 and args[1] == type(None):
                return _py_type_to_arrow_type(args[0], field)
    elif sys.version_info >= (3, 10) and isinstance(field.annotation, types.UnionType):
        args = field.annotation.__args__
        if len(args) == 2:
            for typ in args:
                if typ == type(None):
                    continue
                return _py_type_to_arrow_type(typ, field)
    elif inspect.isclass(field.annotation):
        if issubclass(field.annotation, pydantic.BaseModel):
            # Struct
            fields = _pydantic_model_to_fields(field.annotation)
            return pa.struct(fields)
    #         elif issubclass(field.annotation, FixedSizeListMixin):
    #             return pa.list_(field.annotation.value_arrow_type(), field.annotation.dim())
    return _py_type_to_arrow_type(field.annotation, field)


def _pydantic_to_field(name: str, field) -> pa.Field:
    """Convert a Pydantic field to a PyArrow Field."""
    dt = _pydantic_to_arrow_type(field)
    return pa.field(name, dt, is_nullable(field))


def pydantic_to_arrow_schema(
    model: pydantic.BaseModel, metadata: dict = None
) -> typing.List[pa.Field]:
    """
    convert a pydantic schema to arrow schema in some sort of opinionated way e.g. dealing with complex types
    """
    fields = [
        _pydantic_to_field(name, field) for name, field in model.model_fields.items()
    ]

    schema = pa.schema(fields)

    if metadata:
        schema = schema.with_metadata(metadata)

    return schema



from . import dates, env
from loguru import logger
import os
from glob import glob
import json
from ast import literal_eval


os.environ["LOGURU_LEVEL"] = "DEBUG"

def ingest_files(directory, name, namespace='public',provider=None):
    """
    a very lazy file ingester
    """
    from funkyprompt.core import AbstractContentModel
    from tqdm import tqdm
    from funkyprompt.services import entity_store
    
    M = AbstractContentModel.create_model(name=name, namespace=namespace)
    try:
        M._register()
    except:
        pass
    provider = provider or entity_store
    store = provider(M)
    
    records= []
    for file in tqdm(glob(f"{directory.rstrip('/')}/*.*")):
        with open(file) as f:
            data = f.read()
            """todo this is a temp hack because we have not thought out how we want to deal with entity names for graphs"""
            file = file.split('/')[-1].replace(' ','').replace(',','').split('.')[0]
            
            if len(data):
                records.append(M(name=file, content=data))
    store.update_records(records)

    

def help(question:str, raw_search:bool=False, provider=None):
    """
    ask a question about the codebase
    """
    from funkyprompt.services.models import language_model_client_from_context
    from funkyprompt.services import entity_store
    from funkyprompt.core import AbstractContentModel
    from funkyprompt.core.agents import CallingContext, LanguageModel, MessageStack
    
    lm_client: LanguageModel = language_model_client_from_context(None)
    
    class CodebaseHelpModel(AbstractContentModel):
        class Config:
            name: str = 'codebase'
            namespace: str = 'core'
            description: str = f"Please answer questions about the funkyprompt codebase using the search results you are provided"
            
    provider = provider or entity_store
    result =  provider(CodebaseHelpModel).ask(question)
    
    if raw_search:
        return result
    
    messages= MessageStack.from_q_and_a(question, result)
    
    #return messages

    response = lm_client(messages=messages, functions=None, context=None)
       
    return response

def index_codebase(include_api_json=True, provider = None):
    """
    add all the code into a vector store in a crude sort of way
    """
    
    from glob import glob
    from funkyprompt.services import entity_store
    from funkyprompt.core import AbstractContentModel
    from funkyprompt.core.utils.env import get_repo_root

    provider = provider or entity_store
    def split_string_into_chunks(string, chunk_size=20000):
        """simple chunker"""
        return [string[i : i + chunk_size] for i in range(0, len(string), chunk_size)]

    R = get_repo_root()
    """by using the abstract model which is an entity we are saying the name is unique when adding records
       think about this - its also a problem that the name can be a path or anything
       but the graph node expect some reasonably friendly names without funny characters
       
       TODO: decide what to do and make graph names regex safe
       Another factor is if we want to default to vector search / tabular search or allow entities
       In this case the entity model would not make sense so maybe the right superclass helps here
    """
    M = AbstractContentModel.create_model(name='codebase', namespace='core')
    
    py_files = glob(f"{R}/**/*.py", recursive=True)
    readme_files = glob(f"{R}/**/*.md", recursive=True)
    yaml_files = glob(f"{R}/**/*.yaml", recursive=True)
    records = []

    for file_path in yaml_files:
        with open(file_path, "r") as f:
            f = f.read()
            f = f"""
            ```yaml
            # {file_path}
            {f}
            ```"""

            records.append(M(name=file_path, content=f))
    # Print the list of .py files
    for file_path in py_files:
        with open(file_path, "r") as f:
            f = f.read()
            for i, f in enumerate(split_string_into_chunks(f)):
                f = f"""
                This is a python file
                
                ```python
                #{file_path}
                {f}
                ```"""
                name = file_path if i == 0 else f"{file_path}_{i}"
                records.append(M(name=name, content=f))

    for file_path in readme_files:
        with open(file_path, "r") as f:
            f = f.read()
            records.append(M(name=file_path, content=f))

    hidden_files = []
    for root, dirs, files in os.walk(R):
        # Filter out files that start with a dot
        hidden_files.extend(
            [os.path.join(root, f) for f in files if f.startswith(".") and ".md" in f]
        )

    for file_path in hidden_files:
        with open(file_path, "r") as f:
            f = f.read()
            records.append(M(name=file_path, content=f))

    """
    add records to the store 
    """

    provider(M).update_records(records)
    
    logger.info(f'Done - added {len(records)} records')

     
def coerce_list(o):
    """simple coerce"""
    if isinstance(o,list):
        return o
    if isinstance(o,str):
        try:
            return literal_eval(o)
        except:
            return o.split(',')
        
    raise ValueError(o)
        
def coerce_json(o):
    """simple coerce"""
    if isinstance(o,dict):
        return o
    if isinstance(o,str):
        try:
            return json.loads(o)
        except:
            return literal_eval(o)
        
    raise ValueError(f"{o} must be a string or parseable to json but got {type(o)} that we can not parse")
            